try:
  import cupy as cp
except ModuleNotFoundError:
  raise ModuleNotFoundError("cupy is not installed, please visit https://pypi.org/project/cupy/")

from . import kmeans
from . import kernels

from .PQ import PQ
from .SQ import SQ
from .IVFPQBase import IVFPQBase
from .IVFPQTopk import IVFPQTopk

from .IVFPQ import IVFPQ
from .IVFPQR import IVFPQR



import torch
import cupy as cp
import numpy as np
import math
from .CustomKernel import CustomKernel
from torchpq.util import get_absolute_path

class GetWriteAddressCUDA(CustomKernel):
  def __init__(
      self,
      tpb=256,
      sm_size=48*256*4,
    ):
    super(GetWriteAddressCUDA, self).__init__()
    self.tpb = tpb
    self.sm_size = sm_size

    with open(get_absolute_path("kernels", "GetWriteAddressKernel.cu"), "r") as f:
      self.kernel = f.read()
      
    kernel = (self.kernel
      .replace("_TPB_", str(tpb))
    )

    self.fn = cp.RawKernel(
      kernel,
      'get_write_address',
      backend='nvcc',
      # options=('--maxrregcount=255',),
    )

  def __call__(self, empty_adr, div_of_empty_adr, labels, ioa):
    """
      empty_adr: [n_empty]
      div_of_empty_adr: [n_empty]
      labels: [n_labels] 
      ioa: [n_labels] index of appearance of unique label
    """
    n_empty = empty_adr.shape[0]
    n_labels = labels.shape[0]
    
    write_adr = torch.zeros_like(labels) - 1
    threads_per_block = (self.tpb,)
    blocks_per_grid = (math.ceil(n_labels/self.tpb), )

    # torch.cuda.synchronize()
    self.fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      args=[
        empty_adr.data_ptr(),
        div_of_empty_adr.data_ptr(),
        labels.data_ptr(),
        ioa.data_ptr(),
        write_adr.data_ptr(),
        n_empty, n_labels
      ],
      stream=self.stream
    )
    return write_adr


from .CustomKernel import CustomKernel
from .CustomKernel import Stream

from .ComputeProductCUDA import ComputeProductCUDA
from .GetAddressOfIDCUDA import GetAddressOfIDCUDA
from .GetDivOfAddressCUDA import GetDivOfAddressCUDA
from .GetIOACUDA import GetIOACUDA
from .GetWriteAddressCUDA import GetWriteAddressCUDA
from .PQDecodeCUDA import PQDecodeCUDA


import torch
import cupy as cp
import numpy as np
import math
from .CustomKernel import CustomKernel
from torchpq.util import get_absolute_path

class PQDecodeCUDA(CustomKernel):
  def __init__(
      self,
      tm=2,
      td=8,
    ):
    super(PQDecodeCUDA, self).__init__()
    self.tm = tm
    self.td = td
    self.tpb = 256
    self.sm_size = td * tm * 256 * 4

    with open(get_absolute_path("kernels", "PQDecodeKernel.cu"), "r") as f:
      self.kernel = f.read()

    kernel = (self.kernel
      .replace("_TD_", str(td))
      .replace("_TM_", str(tm))
      .replace("_TPB_", str(self.tpb))
    )

    self.fn = cp.RawKernel(
      kernel,
      'pq_decode',
    )

    self.fn.max_dynamic_shared_size_bytes = self.sm_size

  def __call__(self, codebook, code):
    """
      codebook: torch.Tensor, shape : [m, d, k], dtype : float32
      code: torch.Tensor, shape : [m, n], dtype : uint8
      return: torch.Tensor, shape : [m, d, n], dtype : float32
    """
    m, d, k = codebook.shape
    m, n = code.shape
    assert codebook.shape[0] == code.shape[0]
    assert k == 256

    result = torch.ones(m, d, n, device="cuda:0", dtype=torch.float)

    threads_per_block = (self.tpb,)
    blocks_per_grid = (math.ceil(m/self.tm), math.ceil(d/self.td))
    self.fn(
      grid=blocks_per_grid,
      block=threads_per_block,
      shared_mem = self.sm_size,
      args=[
        codebook.data_ptr(),
        code.data_ptr(),
        result.data_ptr(),
        m, d, n
        ],
      stream=self.stream
    )
    result = result.reshape(m*d, n)
    return result


import torch
import torch.nn as nn
import numpy as np

from .kernels import GetAddressOfIDCUDA
from .kernels import GetDivOfAddressCUDA
from .kernels import GetIOACUDA
from .kernels import GetWriteAddressCUDA
from .SQ import SQ

class IVFPQBase(nn.Module):
  def __init__(
    self,
    d_vector,
    code_size,
    n_cq_clusters,
    blocksize,
    verbose,
    distance,
    cpu_quantizer,
    device,
    ):
    """
    this is a base class for IVFPQ variants

    Parameters:
      d_vector: int
        dimentionality of vectors to be quantized.

      code_size: int, default : 8
        byte size of quantized code, must to be a multiple of 4
        maximum possible code_size depends on GPU architecture
          GPU Architecture:   fp32  fp16 
          Ampere (GA100)      163   326  
          Turing (TU102 etc.) 64    128  
          Volta (GV100)       96    192  
          Pascal and before   48    96   
      
      n_cq_clusters: int, default : 128
        number coarse quantizer clusters
        recommended value is between 4*sqrt(n_data) ~ 16*sqrt(n_data)

      blocksize: int, default : 64
        number of vectors that can be asigned to each cluster of coarse_quantizer initially
        can be expanded using .expand methods, .add method will automatically cal .expand if necessary
        recommended value is (n_data_to_be_stored / n_cq_clusters)

      verbose: int, default : 0
        verbosity

      distance: str, default : 'euclidean'
        type of distance metric
        can be one of: ['euclidean', 'cosine', 'inner']

      cpu_quantizer: SQ or None, default : None
        scalar quantizer used to quantize datapoints to be stored in CPU RAM
        if None is given, CPU RAM will not be used.

      device: str, default : 'cuda:0'
        which device to use
        can be one of: ['cpu', 'cuda', 'cuda:X'] *X represents cuda device index
      
    """
    super(IVFPQBase, self).__init__()
    self.n_cs = 4
    assert code_size % self.n_cs == 0, f"code_size needs to be a multiple of {self.n_cs}"
    
    self.d_vector = d_vector
    self.code_size = code_size
    self.n_cq_clusters = n_cq_clusters
    self.blocksize = blocksize
    self.verbose = verbose
    self.register_buffer("_is_trained", torch.tensor(False))
    self.device = device
    self.cpu_quantizer = cpu_quantizer
    if cpu_quantizer is not None:
      if cpu_quantizer.bits in [32, 16, 8]:
        self.cpu_code_size = d_vector
      elif cpu_quantizer.bits in [4]:
        self.cpu_code_size = d_vector // 2

      if cpu_quantizer.bits == 32:
        self.cpu_dtype = torch.float32
      elif cpu_quantizer.bits == 16:
        self.cpu_dtype = torch.float16
      elif cpu_quantizer.bits in [8, 4]:
        self.cpu_dtype = torch.uint8

    if distance.lower() in ["euclidean", "l2"]:
      distance = "euclidean"
    elif distance.lower() in ["dot", "inner"]:
      distance = "inner"
    elif distance.lower() in ["cosine"]:
      distance = "cosine"
    self.distance = distance
      
    

    storage = torch.zeros(
      code_size // self.n_cs,
      blocksize * n_cq_clusters,
      self.n_cs,
      device=self.device,
      dtype=torch.uint8,
    )

    if cpu_quantizer is None:
      cpu_storage = None
    else:
      cpu_storage = torch.zeros(
        self.cpu_code_size,
        blocksize * n_cq_clusters,
        device="cpu",
        dtype=self.cpu_dtype,
      )

    address2id = torch.ones(
      blocksize * n_cq_clusters,
      device=self.device,
      dtype=torch.long,
    ) * -1
    is_empty = torch.ones(
      blocksize * n_cq_clusters,
      device=self.device,
      dtype=torch.uint8
    )
    div_start = torch.arange(
      n_cq_clusters,
      device=self.device
    ) * blocksize
    div_capacity = torch.ones(
      n_cq_clusters,
      device=self.device,
      dtype=torch.long,
    ) * blocksize
    div_size = torch.zeros(
      n_cq_clusters,
      device=self.device,
      dtype=torch.long
    )
    self.register_buffer("storage", storage)
    self.register_buffer("cpu_storage", cpu_storage)
    self.register_buffer("div_start", div_start)
    self.register_buffer("div_capacity",div_capacity)
    self.register_buffer("div_size", div_size)
    self.register_buffer("address2id", address2id)
    self.register_buffer("is_empty", is_empty)

    self._get_address_of_id_cuda = GetAddressOfIDCUDA(
      tpb=256,
    )
    self._get_div_of_address_cuda = GetDivOfAddressCUDA(
      ta=1,
      tpb=256,
    )
    self._get_ioa_cuda = GetIOACUDA(
      tpb=256,
    )
    self._get_write_address_cuda = GetWriteAddressCUDA(
      tpb=256,
    )

  def __getitem__(self, ids):
    """
      ids: int, or an iterable of integers
    """
    if type(ids) is not torch.Tensor:
      ids = torch.LongTensor(ids)
    if len(ids.shape) == 0:
      ids = ids[None]
    elif len(ids.shape) > 1:
      raise ValueError("only 1D array/tensors are supported")
    return self.get_data_of_id(ids)

  def __repr__(self):
    def get_bytesize_str(bytesize):
      if bytesize < 1024*1:
        bytesize = f"{bytesize} B"
      elif bytesize < 1024**2:
        bytesize = f"{round(bytesize / 1024, 2)} KB"
      elif bytesize < 1024**3:
        bytesize = f"{round(bytesize / 1024**2, 2)} MB"
      elif bytesize < 1024**4:
        bytesize = f"{round(bytesize / 1024**3, 2)} GB"
      elif bytesize < 1024**5:
        bytesize = f"{round(bytesize / 1024**4, 2)} TB"
      return bytesize

    bytesize = self.bytesize
    cpu_bytesize = get_bytesize_str(bytesize[0])
    gpu_bytesize = get_bytesize_str(bytesize[1])

    txt =  f"""
    {type(self).__name__}(
      number of stored items = {self.tot_size},
      total capacity of storage = {self.tot_capacity},
      byte size on CPU = {cpu_bytesize},
      byte size on GPU = {gpu_bytesize},
      code size = {self.code_size} B,
      is trained = {self._is_trained.item()}
    )
    """
    txt = '\n'.join([i[4:] for i in txt.split('\n')])
    return txt

  def __len__(self):
    return self.tot_size

  @staticmethod
  def get_cc():
    if torch.cuda.is_available():
      device_id = torch.cuda.current_device()
      gpu_properties = torch.cuda.get_device_properties(device_id)
      result = (gpu_properties.major, gpu_properties.minor)
    else:
      result = (10, 0)
    return result
  
  @staticmethod
  def normalize(a, dim=-1):
    """
      normalize given tensor along given dimention
    """
    a_norm = a.norm(dim=dim, keepdim=True) + 1e-9
    return a / a_norm

  @property
  def tot_size(self):
    """
      total number of items stored
    """
    return self.div_size.sum().item()

  @property
  def tot_capacity(self):
    """
      total capacity of storage
    """
    return self.div_capacity.sum().item()

  @property
  def bytesize(self):
    """
      the bytesize of cpu and gpu memory that are being used by module
      returns: (cpu_bytesize, gpu_bytesize)
    """
    cpu_bytesize = 0
    gpu_bytesize = 0
    for p in self.state_dict().values():
      if type(p) is torch.Tensor:
        if p.device.type == "cpu":
          cpu_bytesize += p.numel() * p.element_size()
        elif p.device.type == "cuda":
          gpu_bytesize += p.numel() * p.element_size()
    return (cpu_bytesize, gpu_bytesize)

  def _get_address_of_id_old(self, ids):
    batch_size = ids.shape[0]
    address = torch.zeros(batch_size, device=self.device, dtype=torch.long)
    for i in range(batch_size):
      id = ids[i]
      adr = torch.nonzero(self.address2id == id)
      if adr.shape[0] > 0:
        address[i] = adr[0, 0]
      else:
        address[i] = -1
    return address

  def get_address_of_id(self, ids):
    """
      ids: torch.Tensor, shape : [batch_size], dtype : int64
      if there are multiple matching ids, only the address of the first one will be returned
      if there are no matching ids, the address is -1
    """
    if self.address2id.device.type == "cuda":
      return self._get_address_of_id_cuda(self.address2id, ids)
    elif self.address2id.device.type == "cpu":
      return self._get_address_of_id_old(ids)

  def get_id_of_address(self, address):
    """
      address: torch.Tensor, shape : [batch_size], dtype : int64
    """
    return self.address2id[address]

  def _get_div_of_address_old(self, address):
    assert address.max() < self.tot_capacity
    batch_size = address.shape[0]
    div_start = self.div_start
    div_end = div_start + self.div_capacity
    # for adr in address:
    #   mask1 = div_start <= adr
    #   mask2 = div_ends > adr
    #   mask = mask1 & mask2
    #   div = torch.nonzero(mask)
    mask1 = div_start[None, ] <= address[:, None]
    mask2 = div_end[None, ] > address[:, None] # [batch_size, n_cq_clusters]
    mask = mask1 & mask2
    not_found = mask.sum(dim=1) == 0
    mask[not_found, 0] = True
    divs = torch.nonzero(mask)
    divs[not_found, 1] = -1
    return divs[:, 1]

  def get_div_of_address(self, address):
    """
      address: torch.Tensor, shape : [n_address], dtype : int64
      if address is not in range [0, tot_capacity), its div will be -1
    """
    if self.div_start.device.type == "cuda":
      div_start = self.div_start
      div_end = div_start + self.div_capacity
      return self._get_div_of_address_cuda(address, div_start, div_end)
    elif self.div_start.device.type == "cpu":
      return self._get_div_of_address_old(address)

  def get_data_of_address(self, address):
    """
      address: torch.Tensor, shape : [n_address], dtype : int64
      returns: torch.Tensor, shape : [code_size, n_address], dtype : uint8
      address must be in range [0, tot_capacity)
    """
    n_address = address.shape[0]
    assert address.dtype == torch.long
    mask = address < 0
    data = self.storage[:, address] #[code_size//n_cs, n_address, n_cs]
    data = data.transpose(1,2).reshape(self.code_size, n_address)
    if mask.sum() > 0:
      data[:, mask] = 0
    return data

  def get_data_of_id(self, ids):
    """
      ids: torch.Tensor, shape : [n_ids], dtype : int64
      returns: torch.Tensor, [code_size, n_ids]
      if an id doesn't exist, its retrieved data will be zeros.
    """
    address = self.get_address_of_id(ids)
    data = self.get_data_of_address(address)
    return data

  def get_cpu_data_of_address(self, address):
    """
      address: torch.Tensor, shape : [n_address], dtype : int64
      return: torch.Tensor, shape : [cpu_code_size, n_address], dtype : cpu_dtype
      address must be in range [0, tot_capacity)
    """
    if self.cpu_storage is not None:
      n_address = address.shape[0]
      assert address.dtype == torch.long
      mask = address < 0
      data = self.cpu_storage[:, address] #[cpu_code_size, n_address]
      if mask.sum() > 0:
        data[:, mask] = 0
      return data

  def get_cpu_data_of_id(self, ids):
    """
      ids: torch.Tensor, shape : [n_ids], dtype : int64
      returns: torch.Tensor, [cpu_code_size, n_ids]
      if an id doesn't exist, its retrieved data will be zeros.
    """
    address = self.get_address_of_id(ids)
    data = self.get_cpu_data_of_address(address)
    return data

  def set_data_of_address(self, data, address):
    """
      data: torch.Tensor, shape : [code_size // n_cs, n_data, n_cs], dtype : uint8
      address: torch.Tensor, shape : [n_data], dtype : int64
    """
    self.storage[:, address] = data

  def set_data_of_id(self, data, ids):
    """
      data: torch.Tensor, shape : [code_size // n_cs, n_data, n_cs], dtype : uint8
      ids: torch.Tensor, shape : [n_data], dtype : int64
    """
    address = self.get_address_of_id(ids)
    self.set_data_of_address(data, address)

  def set_cpu_data_of_address(self, data, address):
    """
      data: torch.Tensor, shape : [cpu_code_size, n_data, ], dtype : cpu_dtype
      address: torch.Tensor, shape : [n_data], dtype : int64
    """
    if self.cpu_storage is not None:
      assert data.dtype == self.cpu_dtype
      assert data.shape[0] == self.cpu_code_size
      self.cpu_storage[:, address] = data

  def set_cpu_data_of_id(self, data, ids):
    """
      data: torch.Tensor, shape : [code_size // n_cs, n_data, n_cs], dtype : uint8
      ids: torch.Tensor, shape : [n_data], dtype : int64
    """
    if self.cpu_storage is not None:
      address = self.get_address_of_id(ids)
      self.set_cpu_data_of_address(data, address)

  def _get_ioa_old(self, labels, unique_labels=None):
    if unique_labels is None:
      unique_labels = torch.unique(labels) #[n_unique_clusters]
    expanded_labels = labels[:, None].expand(-1, unique_labels.shape[0]) #[n_data, n_unique_clusters]
    mask = expanded_labels == unique_labels[None, :] #[n_data, n_unique_clusters]
    mcs = mask.cumsum(dim=0)
    mcs[[~mask]] = 0
    ioa = mcs.sum(dim=1) - 1

  def get_ioa(self, labels, unique_labels=None):
    if labels.device.type == "cuda":
      return self._get_ioa_cuda(labels, unique_labels)
    elif labels.device.type == "cpu":
      return self._get_ioa_old(labels, unique_labels)

  def _get_write_address_old(self, empty_adr, div_of_empty_adr, labels, ioa):
    write_address = torch.zeros_like(labels)
    for i in range(n_labels):
      div_mask = div_of_empty_adr == labels[i]
      write_adr = empty_adr[div_mask]
      write_adr = write_adr[ioa[i] ]
      write_address[i] = write_adr
    return write_address

  def get_write_address(self, empty_adr, labels, div_of_empty_adr=None, ioa=None):
    if div_of_empty_adr is None:
      div_of_empty_adr = self.get_div_of_address(empty_adr)
    if ioa is None:
      ioa = self.get_ioa(labels)

    if labels.device.type == "cuda":
      return self._get_write_address_cuda(empty_adr, div_of_empty_adr, labels, ioa)
    elif labels.device.type == "cpu":
      return self._get_write_address_old(empty_adr, div_of_empty_adr, labels, ioa)

  def expand(self, clusters):
    """
        clusters: an iterable of integers, len(clusters) <= n_cq_clusters
    """
    if self.verbose > 1:
      print(f"Expanding {len(clusters)} clusters")
    tot = 0
    storage = self.storage
    address2id = self.address2id
    is_empty = self.is_empty
    cpu_storage = self.cpu_storage
    del self.storage, self.address2id, self.is_empty, self.cpu_storage
    for cluster_index in clusters:
      div_start = self.div_start[cluster_index]
      div_cap = self.div_capacity[cluster_index].item()
      div_end = div_start + div_cap

      new_block = torch.zeros(
        self.code_size // self.n_cs,
        div_cap,
        self.n_cs, device=self.device,
        dtype=torch.uint8,
      )
      storage = torch.cat([
        storage[:, :div_end],
        new_block,
        storage[:, div_end:]
      ], dim=1)
      
      if cpu_storage is not None:
        cpu_new_block = torch.zeros(
          self.cpu_code_size,
          div_cap,
          device="cpu",
          dtype=self.cpu_dtype
        )
        cpu_storage = torch.cat([
          cpu_storage[:, :div_end],
          cpu_new_block,
          cpu_storage[:, div_end:]
        ], dim=1)

      new_a2i = torch.ones(div_cap, device=self.device, dtype=address2id.dtype) * -1
      address2id = torch.cat([
        address2id[:div_end],
        new_a2i,
        address2id[div_end:]
      ], dim=0)

      new_is_empty = torch.ones(div_cap, device=self.device, dtype=is_empty.dtype)
      is_empty = torch.cat([
        is_empty[:div_end],
        new_is_empty,
        is_empty[div_end:]
      ], dim=0)

      self.div_capacity[cluster_index] += div_cap
      arange = torch.arange(start=cluster_index+1, end=self.n_cq_clusters)
      self.div_start[arange] += div_cap
      # self.div_start[:] = self.div_capacity.cumsum(dim=0)
      tot += div_cap
    self.register_buffer("storage", storage)
    self.register_buffer("cpu_storage", cpu_storage)
    self.register_buffer("address2id", address2id)
    self.register_buffer("is_empty", is_empty)
    if self.verbose > 1:
      print(f"Storage capacity is increased by {tot}, total capacity: {self.storage.shape[1]}")
  
  def add_to_cpu_ram(self, input, address):
    """
      add scalar quantized input to cpu ram.
    """
    if self.cpu_storage is not None:
      code = self.cpu_storage.encode(input)
      self.set_cpu_data_of_address(code, address)

  def add(self, input, input_ids=None):
    """
      has to be overrided in child class
    """
    raise NotImplementedError("method needs to be overrided in child class")

  def remove(self, remove_ids):
    """
      remove_ids: torch.Tensor, shape : [n_ids], dtype : int64
    """
    assert self._is_trained == True, "Module is not trained"
    n_ids = remove_ids.shape[0]
    address = self.get_address_of_id(remove_ids)
    address = address[address >= 0] #filter out -1 values
    divs = self.get_div_of_address(address)

    self.is_empty[address] = 1
    self.address2id[address] = -1

    unique_divs, counts = divs.unique(return_counts=True)
    self.div_size[unique_divs] -= counts
    if self.verbose > 1:
      print(f"{len(address)} items have been successfully removed, {n_ids - len(address)} items are ignored.")

  def train(self, input, force_retrain=False):
    """
      has to be overrided in child class
    """
    raise NotImplementedError("method needs to be overrided in child class")

  def encode(self, input):
    """
      has to be overrided in child class
    """
    raise NotImplementedError("method needs to be overrided in child class")

  def decode(self, code):
    """
      has to be overrided in child class
    """
    raise NotImplementedError("method needs to be overrided in child class")
    
  def topk(self, query, k, mode=1):
    """
      has to be overrided in child class
    """
    raise NotImplementedError("method needs to be overrided in child class")
    


import torch
import numpy as np

from .IVFPQBase import IVFPQBase
from .kmeans import KMeans
from .PQ import PQ
from .IVFPQTopk import IVFPQTopk

class IVFPQR(IVFPQBase):
  def __init__(
    self,
    d_vector,
    n_subvectors=8,
    n_subvectors_r=8,
    n_cq_clusters=128,
    n_pq_clusters=256,
    blocksize=64,
    verbose=0,
    use_residual=True,
    rerank_factor=2,
    distance="euclidean",
    cpu_quantizer=None,
    device='cuda:0'
    ):
    """
    this is an efficient implementation of IVFPQ+R algorithm
    for more details, you can search 'Searching in one billion vectors: re-rank with source coding'

    Parameters:
      d_vector: int
        dimentionality of vectors to be quantized.

      n_subvectors: int, default : 8
        number of sub-quantizers, needs to be a multiple of 4

      n_subvectors_r: int, default : 8
        number of sub-quantizers for PQ-rerank, need to be a multiple of 4
        maximum number of n_subvectors and n_subvectors_r depends on the GPU architecture
          GPU Architecture   fp32  fp16 
          Ampere (GA100)      163   326  
          Turing (TU102 etc.) 64    128  
          Volta (GV100)       96    192  
          Pascal and before   48    96   
      
      n_cq_clusters: int, default : 128
        number coarse quantizer clusters
        recommended value is between 4*sqrt(n_data) ~ 16*sqrt(n_data)

      n_pq_clusters: int, default : 256
        number of product quantizer clusters
        any value other than 256 is not recommended.

      blocksize: int, default : 64
        number of vectors that can be asigned to each cluster of coarse_quantizer initially
        can be expanded using .expand method, .add method will automatically call .expand if necessary
        recommended value is (n_data_to_be_stored / n_cq_clusters)

      verbose: int, default : 0
        verbosity

      distance: str, default : 'euclidean'
        type of distance metric
        can be one of: ['euclidean', 'cosine', 'inner']

      cpu_quantizer: SQ or None, default : None
        scalar quantizer used to quantize datapoints to be stored in CPU RAM
        if None is given, CPU RAM will not be used.

      device: str, default : 'cuda:0'
        which device to use
        can be one of: ['cpu', 'cuda', 'cuda:X'] *X represents cuda device index
      
    """
    n_cs = 4
    self.n_cs = n_cs
    self.d_vector = d_vector
    self.d_subvector = d_vector // n_subvectors
    self.d_subvector_r = d_vector // n_subvectors_r
    assert n_subvectors % self.n_cs == 0, f"n_subvectors needs to be a multiple of {n_cs}"
    assert n_subvectors_r % self.n_cs == 0, f"n_subvectors_r needs to be a multiple of {n_cs}"
    self.n_subvectors = n_subvectors
    self.n_subvectors_r = n_subvectors_r
    self.n_pq_clusters = n_pq_clusters
    self.use_residual = use_residual
    self.rerank_factor = rerank_factor
    self.n_probe = 1

    cc = self.get_cc()
    if cc[0] < 7 or cc == (7, 2):
      assert n_subvectors <= 48
      assert n_subvectors_r <= 48
    elif cc == (7, 0):
      assert n_subvectors <= 96
      assert n_subvectors_r <= 96
    elif cc == (7, 5):
      assert n_subvectors <= 64
      assert n_subvectors_r <= 64
    elif cc == (8, 0):
      assert n_subvectors <= 163
      assert n_subvectors_r <= 163
    elif cc == (8, 6):
      assert n_subvectors <= 99
      assert n_subvectors_r <= 99
    
    super(IVFPQR, self).__init__(
      d_vector=d_vector,
      code_size=n_subvectors+n_subvectors_r,
      n_cq_clusters=n_cq_clusters,
      blocksize=blocksize,
      verbose=verbose,
      distance=distance,
      cpu_quantizer=cpu_quantizer,
      device=device
    )

    self.coarse_q = KMeans(
      max_iter=25,
      n_clusters=n_cq_clusters,
      distance="inner" if distance == "cosine" else distance,
      init_mode="random",
      verbose=verbose,
    )

    self.product_q = PQ(
      d_vector=d_vector,
      n_subvectors=n_subvectors,
      n_clusters=n_pq_clusters,
      distance="inner" if distance == "cosine" else distance,
      verbose=verbose,
    )
    
    self.product_q_r = PQ(
      d_vector=d_vector,
      n_subvectors=n_subvectors_r,
      n_clusters=n_pq_clusters,
      distance="inner" if distance == "cosine" else distance,
      verbose=verbose,
    )
    self._topk_fn = IVFPQTopk(
      n_subvectors=n_subvectors,
      n_clusters=n_pq_clusters,
      n_cs=n_cs
    )
    self._topk_r_fn = IVFPQTopk(
      n_subvectors=n_subvectors_r,
      n_clusters=n_pq_clusters,
      n_cs=n_cs
    )
  
  def set_coarse_q_max_iter(self, value):
    self.coarse_q.max_iter = value
  
  def set_product_q_max_iter(self, value):
    self.product_q.kmeans.max_iter = value

  def set_product_q_r_max_iter(self, value):
    self.product_q_r.kmeans.max_iter = value
  
  def set_coarse_q_n_redo(self, value):
    self.coarse_q.n_redo = value
  
  def set_product_q_n_redo(self, value):
    self.product_q.kmeans.n_redo = value

  def set_product_q_r_n_redo(self, value):
    self.product_q_r.kmeans.n_redo = value

  def get_data_of_address(self, address):
    """
      address: torch.Tensor, shape : [n_address], dtype : int64
      returns: torch.Tensor, shape : [(n_subvectors+n_subvectors_r), n_address], dtype : uint8
      address must be in range [0, tot_capacity)
    """
    assert address.dtype == torch.long
    n_address = address.shape[0]
    mask = address < 0
    data = self.storage[:, address] #[(n_subvectors+n_subvectors_r)//n_cs, n_address, n_cs]
    data = data.transpose(1,2).reshape(self.n_subvectors + self.n_subvectors_r, n_address)
    if mask.sum() > 0:
      data[:, mask] = 0
    return data
  
  def get_data_of_id(self, ids):
    """
      ids: torch.Tensor, shape : [n_ids], dtype : int64
      returns: torch.Tensor, [(n_subvectors+n_subvectors_r), n_ids]
      if an id doesn't exist, its retrieved data will be zeros.
    """
    address = self.get_address_of_id(ids)
    data = self.get_data_of_address(address)
    return data

  def set_data_of_address(self, data, address):
    """
      data: torch.Tensor, shape : [(n_subvectors+n_subvectors_r) // n_cs, n_data, n_cs], dtype : uint8
      address: torch.Tensor, shape : [n_data], dtype : int64
    """
    self.storage[:, address] = data

  def set_data_of_id(self, data, ids):
    """
      data: torch.Tensor, shape : [(n_subvectors, n_subvectors_r) // n_cs, n_data, n_cs], dtype : uint8
      ids: torch.Tensor, shape : [n_data], dtype : int64
    """
    address = self.get_address_of_id(ids)
    set_data_of_address(data, address)

  def add(self, input, input_ids=None):
    """
      input: torch.Tensor, shape : [n_data, d_vector], dtype : float32
      input_ids: torch.Tensor, shape : [n_data], dtype : int64
    """
    assert self._is_trained == True, "Module is not trained"
    if self.distance == "cosine":
      input = self.normalize(input)
    input = input.contiguous()

    d_vector, n_data = input.shape
    assert d_vector == self.d_vector
    labels = self.coarse_q.predict(input)

    unique_labels, counts = torch.unique(labels, return_counts=True)
    ioa = self.get_ioa(labels, unique_labels)

    # expand storage if necessary
    while True:
      free_space = self.div_capacity[labels] - self.div_size[labels] - (ioa + 1)
      expansion_required = labels[free_space < 0].unique()
      if expansion_required.shape[0] == 0:
        break
      self.expand(expansion_required)

    #get write address
    empty_adr = torch.nonzero(self.is_empty == 1)[:, 0] #[n_empty]
    div_of_empty_adr = self.get_div_of_address(empty_adr) #[n_empty]
    write_address = self.get_write_address(
      empty_adr=empty_adr,
      labels=labels,
      ioa=ioa
    )
    
    # quantize input and store
    code = self.encode(input) #[(n_subvectors+n_subvectors_r), n_data]
    code = code.reshape(
      (self.n_subvectors + self.n_subvectors_r) // self.n_cs,
      self.n_cs,
      n_data
    ).transpose(1, 2)
    self.set_data_of_address(code, write_address)
    if self.cpu_quantizer is not None:
      self.add_to_cpu_ram(input, write_address)

    #store ids
    if input_ids is None:
      if self.tot_size == 0:
        max_id = 0
      else:
        max_id = self.address2id.max()
      input_ids = torch.arange(start=max_id, end=n_data, device=input.device)
    self.address2id[write_address] = input_ids
    self.is_empty[write_address] = 0

    # update number of stored items in each division
    self.div_size[unique_labels] += counts

    if self.verbose > 1:
      print(f"{n_data} new items added")

    return input_ids

  def train(self, input, force_retrain=False):
    """
      input: torch.Tensor, shape : [d_vector, n_data], dtype : float32
      force_retrain: bool, default : False
    """
    if self._is_trained == True and not force_retrain:
      print("Module is already trained")
      return
    d_vector, n_data = input.shape
    assert d_vector == self.d_vector
    if self.distance == "cosine":
      input = self.normalize(input)

    if self.cpu_quantizer is not None:
      if self.verbose > 0:
        print("Start training scalar quantizer")
      self.cpu_quantizer.train(input)

    if self.verbose > 0:
      print("Start training coarse quantizer...")
    self.coarse_q.fit(input)

    if self.verbose > 0:
      print("Start training product quantizer...")
    code = self.product_q.train_codebook(input)

    if self.use_residual:
      reconstructed = self.product_q.decode(code)
      input = input - reconstructed
      if self.verbose > 0:
        print("Start training PQ-rerank with residuals...")
    else:
      if self.verbose > 0:
        print("Start training PQ-rerank...")

    self.product_q_r.train_codebook(input)

    self._is_trained.data = torch.tensor(True)
    if self.verbose > 0:
      print("Successfully trained!")

  def encode(self, input):
    """
      input: torch.Tensor, shape : [d_vector, n_data], dtype : float32
      return: torch.Tensor, shape : [(n_subvectors + n_subvectors_r), n_data], dtype : uint8
    """
    assert self._is_trained == True, "Module is not trained"
    assert input.shape[0] == self.d_vector
    if self.distance == "cosine":
      input = self.normalize(input)
    
    code1 = self.product_q.encode(input) #[n_subvectors, n_data]
    if self.use_residual:
      reconstructed = self.product_q.decode(code1)
      input = input - reconstructed
    code2 = self.product_q_r.encode(input) #[n_subvectors_r, n_data]
    code = torch.cat([code1, code2], dim=0) #[n_subvectors + n_subvectors_r, n_data]
    return code.byte()

  def decode(self, code):
    """
      code: torch.Tensor, shape : [(n_subvectors+n_subvectors_r), n_data], dtype : uint8
      return: torch.Tensor, shape : [d_vector, n_data], dtype : float32
    """
    assert self._is_trained == True, "Module is not trained"
    assert code.shape[0] == self.n_subvectors + self.n_subvectors_r
    code1 = code[:self.n_subvectors, :] #[n_subvectors, n_data]
    code2 = code[self.n_subvectors:, :] #[n_subvectors_r, n_data]
    recon2 = self.product_q_r.decode(code2)
    if self.use_residual:
      recon1 = self.product_q.decode(code1)
      recon2 = recon2 + recon1
    return recon2
  
  @staticmethod
  def get_product(data, precomputed):
      """
        data: torch.Tensor, shape : [n_subvectors_r // n_cs, n_data, n_cs]
        precomputed: torch.Tensor, shape : [n_subvectors_r, n_query, n_pq_clusters]
      """
      o, n_data, n_cs = data.shape
      n_subvectors_r, n_query, n_pq_clusters = precomputed
      assert n_subvectors_r == o * n_cs

      arange = torch.arange(n_subvectors_r, device="cuda:0")
      data = data.transpose(1, 2).reshape(n_subvectors_r, n_data) #[n_subvectors_r, n_data]
      result = precomputed[arange, :, data[:].long() ].sum(dim=1).T #[]
      return result
  
  def topk(self, query, k, mode=2, return_address=False):
    """
      query: torch.Tensor, shape : [d_vector, n_query], dtype : float32
      k: int
      mode: int, default : 2
        mode 2 is generally faster, but MIGHT have slight errors
      return_address: bool, default : False 
    """
    assert self._is_trained == True, "module is not trained"
    d_vector, n_query = query.shape
    assert d_vector == self.d_vector
    assert 0 < k <= self.tot_size

    # select closest cells with Coarse Q
    centroids = self.coarse_q.centroids
    sims = self.coarse_q.sim(query, centroids)
    _, topk_labels = sims.topk(k=self.n_probe, dim=1)
    # topk_labels = torch.sort(topk_labels, dim=1)[0]
    div_start = self.div_start[topk_labels].int()
    if mode == 1: div_size = self.div_capacity[topk_labels].int()
    elif mode == 2: div_size = self.div_size[topk_labels].int()

    # selection with PQ
    reshaped_query = query.reshape(self.n_subvectors, self.d_subvector, n_query)
    codebook = self.product_q.codebook #[n_subvectors, d_subvector, n_pq_clusters]
    precomputed = self.product_q.kmeans.sim(reshaped_query, codebook, normalize=False)#[n_subvectors, n_query, n_pq_clusters]

    selectedv, selected_address = self._topk_fn(
      k=k * self.rerank_factor,
      data=self.storage[:(self.n_subvectors//self.n_cs)],
      precomputed=precomputed,
      is_empty=self.is_empty,
      div_start=div_start,
      div_size=div_size,
    ) #[n_query, k * rerank_factor]
  
    # rerank with PQ-R
    selected_address = selected_address.long()
    codebook2 = self.product_q_r.codebook

    if self.use_residual:
      selected_code = self.storage[:, selected_address, :] #[(code_size)//n_cs, n_query, k*rerank+factor, n_cs]
      selected_code = (
        selected_code
        .transpose(3, 2)
        .transpose(2, 1)
        .reshape(self.code_size, n_query * k*self.rerank_factor)
      )
      selected_code1 = selected_code[:self.n_subvectors, :] #[n_subvectors, n_query*k*rerank_factor]
      selected_recon = (
        self.product_q
        .decode(selected_code1)
        .reshape(d_vector, n_query, k*self.rerank_factor)
        .transpose(0, 1)
      )

      selected_code2 = selected_code[self.n_subvectors:, :] #[n_subvectors_r, n_query*k*rerank_factor]
      selected_residual = (
        self.product_q_r
        .decode(selected_code2)
        .reshape(d_vector, n_query, k*self.rerank_factor)
        .transpose(0, 1)
      )
      selected_recon.add_(selected_residual) #[n_query, d_vector, k*rerank_factor]

      sims = self.product_q_r.kmeans.sim(
        selected_recon,
        query.T[:, :, None], 
        normalize=False
      ).squeeze(2)
      topkv, topk_selected_address = sims.topk(k=k, dim=-1)
      topk_address = torch.gather(selected_address, index=topk_selected_address.long(), dim=1)
      topki = self.get_id_of_address(topk_address)

    else:
      selected_code = self.storage[(self.n_subvectors // self.n_cs):, selected_address, :] #[n_subvectors_r//n_cs, n_query, k*rerank+factor, n_cs]
      selected_code = selected_code.reshape(
        self.n_subvectors_r // self.n_cs,
        n_query * k * self.rerank_factor,
        self.n_cs
      )
      reshaped_query2 = query.reshape(self.n_subvectors_r, self.d_subvector_r, n_query)
      precomputed2 = self.product_q_r.kmeans.sim(reshaped_query2, codebook2, normalize=False)

      is_empty2 = torch.zeros(
        n_query * k * self.rerank_factor,
        device=self.device,
        dtype=torch.uint8
      )
      div_start2 = torch.arange(
        n_query,
        device=self.device,
        dtype=torch.int32
      )[:, None] * k * self.rerank_factor
      div_size2 = torch.empty_like(div_start2)
      div_size2.fill_(k * self.rerank_factor)

      topkv, topk_selected_address = self._topk_r_fn(
        k=k,
        data=selected_code,
        precomputed=precomputed2,
        is_empty=is_empty2,
        div_start=div_start2,
        div_size=div_size2,
      ) #[n_query, k]

      topk_selected_address = topk_selected_address - div_start2
      topk_address = torch.gather(selected_address, index=topk_selected_address.long(), dim=1)
      topki = self.get_id_of_address(topk_address)

    if return_address:
      return (topkv, topki, topk_address)
    else:
      return (topkv, topki)


import torch
import torch.nn as nn

class SQ(nn.Module):
  """
  scalar qantizer
  Parameters:
    bits: int, default : 8
      desired bit size of scalars, can be one of 4, 8, 16, 32
    
    mode: str, default : minmax
      mode of uniform scalar quantization, supported modes are "minmax" and "meanstd"

    alpha: float, default : 1.0
      used in "meanstd"
    
  """
  def __init__(
    self,
    bits=8,
    alpha=1.0,
    mode="minmax",
    ):
    super(SQ, self).__init__()
    self.bits = bits
    self.mode = mode
    self.alpha = alpha
    self.n_bins = 2 ** bits
    
    self.register_buffer("lower", None)
    self.register_buffer("upper", None)
    self.register_buffer("binsize", None)
    self.register_buffer("range", None)
    self.register_buffer("is_trained", torch.tensor(False))


  def _train_minmax(self, input):
    mins = input.min(dim=-1)[0]
    maxs = input.max(dim=-1)[0]
    self.register_buffer("lower", mins )
    self.register_buffer("upper", maxs )
    self.register_buffer("range", (self.upper - self.lower + 1e-8))

  def _train_meanstd(self, input):
    means = input.mean(dim=-1)
    stds = input.std(dim=-1)
    self.register_buffer("lower", means - self.alpha * stds )
    self.register_buffer("upper", means + self.alpha * stds )
    self.register_buffer("range", (self.upper - self.lower + 1e-8))
  
  def train(self, input):
    """
      Train scalar quantizer with input data statistics
      input: torch.Tensor, shape : [d_vector, n_data], dtype : float32
    """
    if self.bits == 4:
      assert input.shape[0] % 2 == 0, "d_vector needs to be divisible by 2"

    if self.bits <= 8:
      if self.mode == "minmax":
        self._train_minmax(input)

      elif self.mode == "meanstd":
        self._train_meanstd(input)
      
    self.is_trained.data = torch.tensor(True)

  def _encode_32bit(self, input):
    return input

  def _decode_32bit(self, code):
    return code
  
  def _encode_16bit(self, input):
    return input.half()

  def _decode_16bit(self, code):
    return code.float()

  def _encode_8bit(self, input):
    input = input - self.lower[:, None]
    input.div_(self.range[:, None])
    input.mul_(self.n_bins)
    input.round_()
    input.clamp_(min=0, max=self.n_bins - 1)
    code = input.byte()
    return code

  def _decode_8bit(self, code):
    output = code.float()
    output.div_(self.n_bins)
    output.mul_(self.range[:, None])
    output.add_(self.lower[:, None])
    return output

  def _encode_4bit(self, input):
    d_vector = input.shape[0]
    input = input - self.lower[:, None]
    input.div_(self.range[:, None])
    input.mul_(self.n_bins)
    input.round_()
    input.clamp_(min=0, max=self.n_bins - 1)
    first = input[:d_vector // 2,]
    last  = input[d_vector // 2:,]
    code = first * self.n_bins + last
    code = code.byte()
    return code

  def _decode_4bit(self, code):
    code = code.float()
    first = code // self.n_bins
    last = code % self.n_bins
    output = torch.cat([first, last], dim=0)
    del first, last, code
    output.div_(self.n_bins)
    output.mul_(self.range[:, None])
    output.add_(self.lower[:, None])
    return output

  def encode(self, input):
    """
      input: torch.Tensor, shape : [d_vector, n_data], dtype : float32
    """
    assert self.is_trained == True, "SQ is not trained"
    if self.bits == 4:
      return self._encode_4bit(input)
    elif self.bits == 8:
      return self._encode_8bit(input)
    elif self.bits == 16:
      return self._encode_16bit(input)
    elif self.bits == 32:
      return self._encode_32bit(input)

  def decode(self, code):
    """
      code: torch.Tensor, shape : [code_size, n_data], dtype : float32 / float16 / uint8
    """
    assert self.is_trained == True, "SQ is not trained"
    if self.bits == 4:
      return self._decode_4bit(code)
    elif self.bits == 8:
      return self._decode_8bit(code)
    elif self.bits == 16:
      return self._decode_16bit(code)
    elif self.bits == 32:
      return self._decode_32bit(code)